"""
    A module for handle .h5 files for Swift.
"""
from __future__ import annotations

import datetime
import io
import json
import os
import pathlib
import threading
import typing
import uuid

import h5py
import numpy
import numpy.typing

from nion.swift.model import StorageHandler
from nion.swift.model import Utility
from nion.utils import DateTime
from nion.utils import Geometry
from nion.utils import Registry

if typing.TYPE_CHECKING:
    from nion.data import DataAndMetadata


PersistentDictType = typing.Dict[str, typing.Any]
_NDArray = numpy.typing.NDArray[typing.Any]


def make_directory_if_needed(directory_path: str) -> None:
    """
        Make the directory path, if needed.
    """
    if os.path.exists(directory_path):
        if not os.path.isdir(directory_path):
            raise OSError("Path is not a directory:", directory_path)
    else:
        os.makedirs(directory_path)


def get_write_chunk_shape_for_data(data_shape: DataAndMetadata.ShapeType, data_dtype: numpy.typing.DTypeLike) -> typing.Optional[DataAndMetadata.ShapeType]:
    """
    Calculate an appropriate write chunk shape for a given data shape and dtype.

    The target chunk size is 580 kB which seems to be a sweet spot according to benchmarks.
    The algorithm assumes that the data is c-contiguous in memory.

    If the total number of chunks that the calculated chunk shape would lead to is less than 100 (i.e. the file will
    be less than 58 MB in size) or if the data shape is not suitable for chunking, return None.
    """
    data_dtype = numpy.dtype(data_dtype)

    target_chunk_size = 580*1024/data_dtype.itemsize
    chunk_size = 1
    counter = len(data_shape)
    chunk_shape = [1] * len(data_shape)
    while chunk_size < target_chunk_size and counter > 0:
        counter -= 1
        chunk_size *= data_shape[counter]
        chunk_shape[counter] = data_shape[counter]

    if chunk_size == 0: # This means one of the input dimensions was "0", so chunking cannot be used
        return None

    chunk_size //= data_shape[counter]
    remaining_elements = min(max(target_chunk_size // chunk_size, 1), data_shape[counter])
    chunk_shape[counter] = int(remaining_elements)

    n_chunks = 1
    for i in range(len(chunk_shape)):
        n_chunks *= data_shape[i] // chunk_shape[i]
    if n_chunks < 100:
        return None

    return tuple(chunk_shape)


_HDF5FilePointer = typing.Any


class HDF5FileEntry:
    def __init__(self, path: pathlib.Path) -> None:
        self.__lock = threading.RLock()
        self.__path = path
        self.__fp: typing.Optional[_HDF5FilePointer] = None
        self._count = 0

    @property
    def fp(self) -> _HDF5FilePointer:
        with self.__lock:
            self.open()
            assert self.__fp
            return self.__fp

    def open(self) -> None:
        with self.__lock:
            if not self.__fp:
                self.__path.parent.mkdir(parents=True, exist_ok=True)
                self.__fp = h5py.File(self.__path, "a")

    def close(self) -> None:
        with self.__lock:
            if self.__fp:
                self.__fp.close()
            self.__fp = None


class HDF5FileManager:
    def __init__(self) -> None:
        self.__file_entries: typing.Dict[pathlib.Path, HDF5FileEntry] = dict()
        self.__lock = threading.RLock()

    def _clear(self) -> None:
        # for tests only
        self.__file_entries.clear()

    @property
    def _open_count(self) -> int:
        return len(self.__file_entries.items())

    def open(self, path: pathlib.Path) -> HDF5FileEntry:
        with self.__lock:
            if not path in self.__file_entries:
                self.__file_entries[path] = HDF5FileEntry(path)
            self.__file_entries[path]._count += 1
            return self.__file_entries[path]

    def close(self, path: pathlib.Path) -> None:
        with self.__lock:
            self.__file_entries[path]._count -= 1
            if self.__file_entries[path]._count == 0:
                self.__file_entries.pop(path).close()

    def force_close(self, path: pathlib.Path) -> None:
        with self.__lock:
            if path in self.__file_entries:
                self.__file_entries[path].close()


_file_manager = HDF5FileManager()


class HDF5Handler(StorageHandler.StorageHandler):
    count = 0  # useful for detecting leaks in tests
    open = 0  # useful for detecting unclosed files

    def __init__(self, file_path: typing.Union[str, pathlib.Path]) -> None:
        self.__file_path = str(file_path)
        self.__lock = threading.RLock()
        self.__file = _file_manager.open(pathlib.Path(self.__file_path))
        self.__dataset: typing.Any = None
        self._write_count = 0
        HDF5Handler.count += 1

    def close(self) -> None:
        HDF5Handler.count -= 1
        self.__close_fp()
        _file_manager.close(pathlib.Path(self.__file_path))

    @property
    def factory(self) -> StorageHandler.StorageHandlerFactoryLike:
        return HDF5HandlerFactory()

    @property
    def storage_handler_type(self) -> str:
        return "hdf5"

    # called before the file is moved; close but don't count.
    def prepare_move(self) -> None:
        with self.__lock:
            self.__close_fp()
            _file_manager.force_close(pathlib.Path(self.__file_path))

    @property
    def reference(self) -> str:
        return self.__file_path

    @property
    def is_valid(self) -> bool:
        return True

    def get_extension(self) -> str:
        return ".h5"

    def __write_properties_to_dataset(self, properties: PersistentDictType) -> None:
        with self.__lock:
            assert self.__dataset is not None

            class JSONEncoder(json.JSONEncoder):
                def default(self, obj: typing.Any) -> typing.Any:
                    if isinstance(obj, Geometry.IntPoint) or isinstance(obj, Geometry.IntSize) or isinstance(obj, Geometry.IntRect) or isinstance(obj, Geometry.FloatPoint) or isinstance(obj, Geometry.FloatSize) or isinstance(obj, Geometry.FloatRect):
                        return tuple(obj)
                    else:
                        return json.JSONEncoder.default(self, obj)

            json_io = io.StringIO()
            json.dump(Utility.clean_dict(properties), json_io, cls=JSONEncoder)
            json_str = json_io.getvalue()

            self.__dataset.attrs["properties"] = json_str

    def __ensure_dataset(self) -> None:
        with self.__lock:
            if self.__dataset is None:
                if "data" in self.__file.fp:
                    self.__dataset = self.__file.fp["data"]
                else:
                    self.__dataset = self.__file.fp.create_dataset("data", data=numpy.empty((0,)))

    def write_data(self, data: _NDArray, file_datetime: datetime.datetime) -> None:
        with self.__lock:
            assert data is not None
            json_properties = None
            # handle three cases:
            #   1 - 'data' doesn't yet exist (require_dataset)
            #   2 - 'data' exists but is a different size (delete, then require_dataset)
            #   3 - 'data' exists and is the same size (overwrite)
            if not "data" in self.__file.fp:
                # case 1
                chunks = get_write_chunk_shape_for_data(data.shape, data.dtype)
                self.__dataset = self.__file.fp.require_dataset("data", shape=data.shape, dtype=data.dtype, chunks=chunks)
            else:
                if self.__dataset is None:
                    self.__dataset = self.__file.fp["data"]
                if self.__dataset.shape != data.shape or self.__dataset.dtype != data.dtype:
                    # case 2
                    json_properties = self.__dataset.attrs.get("properties", "")
                    self.__close_fp()
                    os.remove(self.__file_path)
                    chunks = get_write_chunk_shape_for_data(data.shape, data.dtype)
                    self.__dataset = self.__file.fp.require_dataset("data", shape=data.shape, dtype=data.dtype, chunks=chunks)
            self.__copy_data(data)
            if json_properties is not None:
                self.__dataset.attrs["properties"] = json_properties
            self.__file.fp.flush()

    def reserve_data(self, data_shape: DataAndMetadata.ShapeType, data_dtype: numpy.typing.DTypeLike, file_datetime: datetime.datetime) -> None:
        # reserve data of the given shape and dtype, filled with zeros
        with self.__lock:
            json_properties = None
            # first read existing properties and then close existing data set and file.
            if "data" in self.__file.fp:
                if self.__dataset is None:
                    self.__dataset = self.__file.fp["data"]
                json_properties = self.__dataset.attrs.get("properties", "")
                self.__close_fp()
                os.remove(self.__file_path)
                self.__file.open()
            # reserve the data
            chunks = get_write_chunk_shape_for_data(data_shape, data_dtype)
            self.__dataset = self.__file.fp.require_dataset("data", shape=data_shape, dtype=data_dtype, fillvalue=0, chunks=chunks)
            if json_properties is not None:
                self.__dataset.attrs["properties"] = json_properties
            self.__file.fp.flush()

    def __copy_data(self, data: _NDArray) -> None:
        if id(data) != id(self.__dataset):
            self.__dataset[:] = data
            self._write_count += 1

    def write_properties(self, properties: PersistentDictType, file_datetime: datetime.datetime) -> None:
        with self.__lock:
            self.__ensure_dataset()
            self.__write_properties_to_dataset(properties)
            self.__file.fp.flush()

    def read_properties(self) -> PersistentDictType:
        with self.__lock:
            self.__ensure_dataset()
            json_properties = self.__dataset.attrs.get("properties", "")
            return typing.cast(PersistentDictType, json.loads(json_properties))

    def read_data(self) -> typing.Optional[_NDArray]:
        with self.__lock:
            self.__ensure_dataset()
            if self.__dataset.shape == (0, ):
                return None
            return typing.cast(typing.Optional[_NDArray], self.__dataset)

    def remove(self) -> None:
        self.__close_fp()
        if os.path.isfile(self.__file_path):
            os.remove(self.__file_path)

    def __close_fp(self) -> None:
        self.__dataset = None
        self.__file.close()


class HDF5HandlerFactory(StorageHandler.StorageHandlerFactoryLike):

    def get_storage_handler_type(self) -> str:
        return "hdf5"

    def is_matching(self, file_path: str) -> bool:
        if file_path.endswith(".h5") and os.path.exists(file_path):
            return True
        return False

    def make(self, file_path: pathlib.Path) -> StorageHandler.StorageHandler:
        return HDF5Handler(self.make_path(file_path))

    def make_path(self, file_path: pathlib.Path) -> str:
        return str(file_path.with_suffix(self.get_extension()))

    def get_extension(self) -> str:
        return ".h5"


class HDFImportExportDriver:
    def __init__(self) -> None:
        pass

    def read_data(self, file_path: pathlib.Path, storage_handler_provider: StorageHandler.StorageHandlerProvider) -> StorageHandler.StorageHandlerImportData:
        storage_handlers = list[StorageHandler.StorageHandler]()
        uuid_map = dict[uuid.UUID, uuid.UUID]()
        items = list[PersistentDictType]()
        fp = h5py.File(file_path, "r")
        if "data" in fp:
            data_group = fp["data"]
            for key in sorted(data_group.keys()):
                ds = data_group[key]
                data_item_uuid = uuid.uuid4()
                data_item_properties = json.loads(ds.attrs["properties"])
                data_item_data = ds
                if "uuid" in data_item_properties:
                    uuid_map[uuid.UUID(data_item_properties["uuid"])] = data_item_uuid
                data_item_properties["uuid"] = str(data_item_uuid)
                storage_handler_attributes = StorageHandler.StorageHandlerAttributes(
                    data_item_uuid,
                    DateTime.utcnow(),
                    data_item_properties.get("session_id", None),
                    ds.nbytes
                )
                storage_handler = storage_handler_provider.make_storage_handler(storage_handler_attributes)
                storage_handler.write_data(data_item_data, storage_handler_attributes.created_local)
                storage_handler.write_properties(data_item_properties, storage_handler_attributes.created_local)
                storage_handlers.append(storage_handler)
        if "index" in fp:
            index_group = fp["index"]
            for key in sorted(index_group.attrs.keys()):
                items.append(json.loads(index_group.attrs[key]))
        fp.close()
        return StorageHandler.StorageHandlerImportData(storage_handlers, uuid_map, items)

    def write_display_item(self, path: pathlib.Path, items: typing.Sequence[StorageHandler.StorageHandlerExportItem]) -> None:
        path.unlink(missing_ok=True)
        fp = h5py.File(path, "a")
        index_group = fp.create_group("index")
        data_group = fp.create_group("data")
        data_index = 0
        for index, item in enumerate(items):
            index_group.attrs["1"] = json.dumps(item.write_to_dict())
            for data_item in item.data_items:
                ds = data_group.create_dataset(str(data_index), data=data_item.data)
                ds.attrs["properties"] = json.dumps(data_item.write_to_dict())
                data_index += 1
        fp.close()


class HDFImportExportDriverFactory:
    driver_id = "nhdf-io-handler"
    title = "NData HDF"
    extensions = ["nhdf"]

    def make_import_export_driver(self) -> HDFImportExportDriver:
        return HDFImportExportDriver()


Registry.register_component(HDFImportExportDriverFactory(), {"import-export-driver-factory"})


"""
Architectural Decision Records.

ADR 2024-02-19. Do not use Single Writer Multiple Reader (SWMR) mode for HDF5 files. This requires the writing process
to set the flag and we don't have control of the writers outside of our process. In addition, the docs curently say that
SWMR is not supported on Windows.
"""



from __future__ import annotations

import json
import typing

from nion.swift.model import Utility
from nion.utils import Event
from nion.utils import Observable


class Feature(Observable.Observable):
    def __init__(self, feature_id: str, description: str, enabled: bool = False) -> None:
        super().__init__()
        self.feature_id = feature_id
        self.description = description
        self.__enabled = enabled

    @property
    def enabled(self) -> bool:
        return self.__enabled

    @enabled.setter
    def enabled(self, enabled: bool) -> None:
        if self.__enabled != enabled:
            self.__enabled = enabled
            self.notify_property_changed("enabled")


class FeatureManager(Observable.Observable, metaclass=Utility.Singleton):
    def __init__(self) -> None:
        super().__init__()
        self.__features = list[Feature]()
        self.__feature_listeners = list[Event.EventListener]()

    @property
    def enabled_feature_str(self) -> str:
        return json.dumps({feature.feature_id: feature.enabled for feature in self.features})

    @enabled_feature_str.setter
    def enabled_feature_str(self, enabled_features_str: str) -> None:
        try:
            enabled_features = json.loads(enabled_features_str)
            for enabled_feature in enabled_features.keys():
                feature = self.get_feature(enabled_feature)
                if feature:
                    feature.enabled = enabled_features[enabled_feature]
        except Exception as e:
            pass  # don't fail to launch due to bad json

    @property
    def features(self) -> typing.Sequence[Feature]:
        return list(self.__features)

    def add_feature(self, feature: Feature) -> None:
        self.__features.append(feature)
        self.__feature_listeners.append(feature.property_changed_event.listen(self.__feature_property_changed))

    def __feature_property_changed(self, property_name: str) -> None:
        if property_name == "enabled":
            self.notify_property_changed("enabled_feature_str")

    def get_feature(self, feature_id: str) -> typing.Optional[Feature]:
        for feature in self.__features:
            if feature.feature_id == feature_id:
                return feature
        return None

    def is_feature_enabled(self, feature_id: str) -> bool:
        feature = self.get_feature(feature_id)
        return feature.enabled if feature else False




from __future__ import annotations

import dataclasses
import datetime
import pathlib
import typing
import uuid

import numpy
import numpy.typing


PersistentDictType = typing.Dict[str, typing.Any]
_NDArray = numpy.typing.NDArray[typing.Any]


class StorageHandlerFactoryLike(typing.Protocol):

    def get_storage_handler_type(self) -> str: ...

    def is_matching(self, file_path: str) -> bool: ...

    def make(self, file_path: pathlib.Path) -> StorageHandler: ...

    def make_path(self, file_path: pathlib.Path) -> str: ...

    def get_extension(self) -> str: ...


class StorageHandler(typing.Protocol):

    def close(self) -> None: ...

    @property
    def factory(self) -> StorageHandlerFactoryLike: raise NotImplementedError()

    @property
    def storage_handler_type(self) -> str: raise NotImplementedError()

    @property
    def reference(self) -> str: raise NotImplementedError()

    @property
    def is_valid(self) -> bool: raise NotImplementedError()

    def read_properties(self) -> PersistentDictType: ...

    def read_data(self) -> typing.Optional[_NDArray]: ...

    def write_properties(self, properties: PersistentDictType, file_datetime: datetime.datetime) -> None: ...

    def write_data(self, data: _NDArray, file_datetime: datetime.datetime) -> None: ...

    def reserve_data(self, data_shape: typing.Tuple[int, ...], data_dtype: numpy.typing.DTypeLike, file_datetime: datetime.datetime) -> None: ...

    def prepare_move(self) -> None: ...

    def remove(self) -> None: ...


@dataclasses.dataclass
class StorageHandlerAttributes:
    """Attributes for organizing within storage system."""
    uuid: uuid.UUID
    created_local: datetime.datetime
    session_id: str | None
    n_bytes: int
    _force_large_format: bool = False


class StorageHandlerProvider(typing.Protocol):
    def make_storage_handler(self, attributes: StorageHandlerAttributes) -> StorageHandler: ...


@dataclasses.dataclass
class StorageHandlerImportData:
    storage_handlers: typing.Sequence[StorageHandler]
    uuid_map: typing.Mapping[uuid.UUID, uuid.UUID]
    items: typing.Sequence[PersistentDictType]


class StorageHandlerExportDataItem(typing.Protocol):
    def write_to_dict(self) -> PersistentDictType: ...

    @property
    def data(self) -> _NDArray: raise NotImplementedError()


class StorageHandlerExportItem(typing.Protocol):
    def write_to_dict(self) -> PersistentDictType: ...

    @property
    def data_items(self) -> typing.Sequence[StorageHandlerExportDataItem]: raise NotImplementedError()



import collections.abc
import typing

# standardized metadata paths, mapping to properties

session_key_map = {
    'stem.session.site': {'paths': ['site'], 'type': 'string'},
    'stem.session.instrument': {'paths': ['instrument'], 'type': 'string'},
    'stem.session.detector': {'paths': ['detector'], 'type': 'string'},
    'stem.session.task': {'paths': ['task'], 'type': 'string'},
    'stem.session.microscopist': {'paths': ['microscopist'], 'type': 'string'},
    'stem.session.sample': {'paths': ['sample'], 'type': 'string'},
    'stem.session.sample_area': {'paths': ['sample_area'], 'type': 'string'},
    'stem.session.label': {'paths': ['label'], 'type': 'string'},
    'stem.session.sample_source': {'paths': ['sample_source'], 'type': 'string'},
    'stem.session.sample_formula': {'paths': ['sample_formula'], 'type': 'string'},
}

# 'hardware_source' should be 'detector' at some point in the future. see HMSA file format for more thoughts.

# keys can exclude the unit suffix if units are SI

key_map = {
    # the info about the primary detector
    'stem.hardware_source.id': {'paths': ['hardware_source.hardware_source_id'], 'type': 'string'},
    'stem.hardware_source.name': {'paths': ['hardware_source.hardware_source_name'], 'type': 'string'},
    'stem.hardware_source.frame_number': {'paths': ['hardware_source.frame_number'], 'type': 'int'},
    'stem.hardware_source.valid_rows': {'paths': ['hardware_source.valid_rows'], 'type': 'int'},

    'stem.signal_type': {'paths': ['hardware_source.signal_type'], 'type': 'string'},
    # EDXS, WDS, ELS, AES, PES, XRF, CLS, GAM, BEI, CBED, EBSD, EDIF, LEED, OPR, OPT, PIXE, RHEED, SEI, SXES, TEM
    # see HMSA

    'stem.high_tension': {'paths': ['instrument.high_tension',
                                    'hardware_source.autostem.high_tension_v'], 'type': 'integer'},
    'stem.high_tension_v': {'paths': ['instrument.high_tension',
                                    'hardware_source.autostem.high_tension_v'], 'type': 'integer'},
    'stem.gun_type': {'paths': ['instrument.gun_type', 'hardware_source.gun_type'], 'type': 'string'},
    'stem.convergence_angle': {'paths': ['instrument.convergence_angle',
                                         'hardware_source.convergence_angle_rad'], 'type': 'real'},
    'stem.convergence_angle_rad': {'paths': ['instrument.convergence_angle',
                                         'hardware_source.convergence_angle_rad'], 'type': 'real'},
    'stem.collection_angle': {'paths': ['instrument.collection_angle',
                                        'hardware_source.collection_angle_rad'], 'type': 'real'},
    'stem.collection_angle_rad': {'paths': ['instrument.collection_angle',
                                        'hardware_source.collection_angle_rad'], 'type': 'real'},
    'stem.probe_size': {'paths': ['instrument.probe_size', 'hardware_source.probe_size_m2'], 'type': 'real'},
    'stem.probe_size_m2': {'paths': ['instrument.probe_size', 'hardware_source.probe_size_m2'], 'type': 'real'},
    'stem.beam_current': {'paths': ['instrument.beam_current', 'hardware_source.beam_current_a'], 'type': 'real'},
    'stem.beam_current_a': {'paths': ['instrument.beam_current', 'hardware_source.beam_current_a'], 'type': 'real'},
    'stem.defocus': {'paths': ['instrument.defocus', 'hardware_source.defocus_m'], 'type': 'real'},
    'stem.defocus_m': {'paths': ['instrument.defocus', 'hardware_source.defocus_m'], 'type': 'real'},

    'stem.eels.spectrum_type': {'paths': ['hardware_source.eels_spectrum_type'], 'type': 'string'},
    'stem.eels.resolution_eV': {'paths': ['hardware_source.eels_resolution_eV'], 'type': 'string'},
    'stem.eels.is_monochromated': {'paths': ['hardware_source.eels_is_monochromated'], 'type': 'boolean'},

    'stem.camera.binning': {'paths': ['hardware_source.binning'], 'type': 'integer'},
    'stem.camera.channel_id': {'paths': ['hardware_source.channel_id'], 'type': 'string'},
    'stem.camera.channel_index': {'paths': ['hardware_source.channel_index'], 'type': 'integer'},
    'stem.camera.channel_name': {'paths': ['hardware_source.channel_name'], 'type': 'string'},
    'stem.camera.exposure': {'paths': ['hardware_source.exposure'], 'type': 'real'},
    'stem.camera.exposure_s': {'paths': ['hardware_source.exposure'], 'type': 'real'},
    'stem.camera.frame_index': {'paths': ['hardware_source.frame_index'], 'type': 'integer'},
    'stem.camera.frame_number': {'paths': ['hardware_source.frame_number'], 'type': 'integer'},
    'stem.camera.valid_rows': {'paths': ['hardware_source.valid_rows'], 'type': 'integer'},
    'stem.camera.detector_current': {'paths': ['hardware_source.detector_current'], 'type': 'real'},

    'stem.scan.center_x_nm': {'paths': ['scan.center_x_nm', 'hardware_source.center_x_nm'], 'type': 'real'},
    'stem.scan.center_y_nm': {'paths': ['scan.center_y_nm', 'hardware_source.center_y_nm'], 'type': 'real'},
    'stem.scan.fov_nm': {'paths': ['scan.fov_nm', 'hardware_source.fov_nm'], 'type': 'real'},
    'stem.scan.rotation': {'paths': ['scan.rotation', 'hardware_source.rotation'], 'type': 'real'},
    'stem.scan.rotation_rad': {'paths': ['scan.rotation', 'hardware_source.rotation'], 'type': 'real'},
    'stem.scan.scan_id': {'paths': ['scan.scan_id', 'hardware_source.scan_id'], 'type': 'string'},
    'stem.scan.valid_rows': {'paths': ['scan.valid_rows', 'hardware_source.valid_rows'], 'type': 'integer'},

    'stem.scan.channel_id': {'paths': ['hardware_source.channel_id'], 'type': 'string'},
    'stem.scan.channel_index': {'paths': ['hardware_source.channel_index'], 'type': 'integer'},
    'stem.scan.channel_name': {'paths': ['hardware_source.channel_name'], 'type': 'string'},
    'stem.scan.frame_time': {'paths': ['hardware_source.exposure'], 'type': 'real'},
    'stem.scan.frame_time_s': {'paths': ['hardware_source.exposure'], 'type': 'real'},
    'stem.scan.frame_index': {'paths': ['hardware_source.frame_index'], 'type': 'integer'},
    'stem.scan.pixel_time_us': {'paths': ['hardware_source.pixel_time_us'], 'type': 'real'},
    'stem.scan.line_time_us': {'paths': ['hardware_source.line_time_us'], 'type': 'real'},
}


def has_metadata_value(metadata_source: typing.Any, key: str) -> bool:
    """Return whether the metadata value for the given key exists.

    There are a set of predefined keys that, when used, will be type checked and be interoperable with other
    applications. Please consult reference documentation for valid keys.

    If using a custom key, we recommend structuring your keys in the '<group>.<attribute>' format followed
    by the predefined keys. e.g. 'session.instrument' or 'camera.binning'.

    Also note that some predefined keys map to the metadata ``dict`` but others do not. For this reason, prefer
    using the ``metadata_value`` methods over directly accessing ``metadata``.
    """
    d: typing.Mapping[str, typing.Any] | None = None

    desc = session_key_map.get(key)
    if desc is not None:
        d = getattr(metadata_source, "session_metadata", None)
        if d is None:
            d = getattr(metadata_source, "metadata", None)
            if d is not None:
                d = d.get("session_metadata", dict())
            elif isinstance(metadata_source, collections.abc.Mapping):
                d = metadata_source.get("session_metadata", dict())
    if d is None:
        desc = key_map.get(key)
        if desc is not None:
            d = getattr(metadata_source, "metadata", metadata_source)
        elif isinstance(metadata_source, collections.abc.Mapping):
            d = metadata_source

    if desc is not None and d is not None:
        for path in desc["paths"]:
            path_components = path.split(".")
            for k in path_components[:-1]:
                d =  d.get(k, dict()) if d is not None else None
            if d is not None:
                return path_components[-1] in d

    return False


def get_metadata_value(metadata_source: typing.Any, key: str) -> typing.Any:
    """Get the metadata value for the given key.

    There are a set of predefined keys that, when used, will be type checked and be interoperable with other
    applications. Please consult reference documentation for valid keys.

    If using a custom key, we recommend structuring your keys in the '<group>.<attribute>' format followed
    by the predefined keys. e.g. 'session.instrument' or 'camera.binning'.

    Also note that some predefined keys map to the metadata ``dict`` but others do not. For this reason, prefer
    using the ``metadata_value`` methods over directly accessing ``metadata``.
    """
    d: typing.Mapping[str, typing.Any] | None = None

    desc = session_key_map.get(key)
    if desc is not None:
        d = getattr(metadata_source, "session_metadata", None)
        if d is None:
            d = getattr(metadata_source, "metadata", None)
            if d is not None:
                d = d.get("session_metadata", dict())
            elif isinstance(metadata_source, collections.abc.Mapping):
                d = metadata_source.get("session_metadata", dict())
    if d is None:
        desc = key_map.get(key)
        if desc is not None:
            d = getattr(metadata_source, "metadata", metadata_source)
        elif isinstance(metadata_source, collections.abc.Mapping):
            d = metadata_source

    if desc is not None and d is not None:
        for path in desc["paths"]:
            path_components = path.split(".")
            for k in path_components:
                d =  d.get(k) if d is not None else None
            if d is not None:
                return d

    return None


def set_metadata_value(metadata_source: typing.Any, key: str, value: typing.Any) -> None:
    """Set the metadata value for the given key.

    There are a set of predefined keys that, when used, will be type checked and be interoperable with other
    applications. Please consult reference documentation for valid keys.

    If using a custom key, we recommend structuring your keys in the '<group>.<attribute>' format followed
    by the predefined keys. e.g. 'session.instrument' or 'camera.binning'.

    Also note that some predefined keys map to the metadata ``dict`` but others do not. For this reason, prefer
    using the ``metadata_value`` methods over directly accessing ``metadata``.
    """
    _set_session_metadata_fn = getattr(metadata_source, "_set_session_metadata", None)
    _set_metadata_fn = getattr(metadata_source, "_set_metadata", None)

    d: typing.MutableMapping[str, typing.Any] | None = None
    path: str | None = None
    is_session_metadata = False
    has_session_metadata = False

    desc = session_key_map.get(key)
    if desc is not None:
        path = desc["paths"][0]
        d = getattr(metadata_source, "session_metadata", None)
        if d is None:
            d = getattr(metadata_source, "metadata", None)
            if d is not None:
                path = "session_metadata." + path
            elif isinstance(metadata_source, collections.abc.MutableMapping):
                path = "session_metadata." + path
                d = metadata_source
        else:
            is_session_metadata = True
    if d is None:
        desc = key_map.get(key)
        if desc is not None:
            path = desc["paths"][0]
            d = getattr(metadata_source, "metadata", metadata_source)
        elif isinstance(metadata_source, collections.abc.MutableMapping):
            d = metadata_source

    changed = False

    if desc is not None and d is not None and path is not None:
        path_components = path.split(".")
        d_ = d
        for k in path_components[:-1]:
            d_ =  d_.setdefault(k, dict()) if d is not None else None
        if d_ is not None:
            d_[path_components[-1]] = value
            changed = True

    if changed and d is not None and path is not None:
        if is_session_metadata:
            assert callable(_set_session_metadata_fn)
            _set_session_metadata_fn(d)
        elif callable(_set_metadata_fn):
            _set_metadata_fn(d)
    else:
        raise KeyError()


def delete_metadata_value(metadata_source: typing.Any, key: str) -> None:
    """Delete the metadata value for the given key.

    There are a set of predefined keys that, when used, will be type checked and be interoperable with other
    applications. Please consult reference documentation for valid keys.

    If using a custom key, we recommend structuring your keys in the '<dotted>.<group>.<attribute>' format followed
    by the predefined keys. e.g. 'stem.session.instrument' or 'stm.camera.binning'.

    Also note that some predefined keys map to the metadata ``dict`` but others do not. For this reason, prefer
    using the ``metadata_value`` methods over directly accessing ``metadata``.
    """
    _set_session_metadata_fn = getattr(metadata_source, "_set_session_metadata", None)
    _set_metadata_fn = getattr(metadata_source, "_set_metadata", None)

    d: typing.MutableMapping[str, typing.Any] | None = None
    path: str | None = None
    is_session_metadata = False
    has_session_metadata = False

    desc = session_key_map.get(key)
    if desc is not None:
        path = desc["paths"][0]
        d = getattr(metadata_source, "session_metadata", None)
        if d is None:
            d = getattr(metadata_source, "metadata", None)
            if d is not None:
                path = "session_metadata." + path
            elif isinstance(metadata_source, collections.abc.MutableMapping):
                path = "session_metadata." + path
                d = metadata_source
        else:
            is_session_metadata = True
    if d is None:
        desc = key_map.get(key)
        if desc is not None:
            path = desc["paths"][0]
            d = getattr(metadata_source, "metadata", metadata_source)
        elif isinstance(metadata_source, collections.abc.MutableMapping):
            d = metadata_source

    changed = False

    if desc is not None and d is not None and path is not None:
        path_components = path.split(".")
        d_ = d
        for k in path_components[:-1]:
            d_ =  d_.get(k, dict()) if d is not None else None
        if d_ is not None and path_components[-1] in d_:
            d_.pop(path_components[-1], None)
            changed = True

    if changed and d is not None and path is not None:
        if is_session_metadata:
            assert callable(_set_session_metadata_fn)
            _set_session_metadata_fn(d)
        elif callable(_set_metadata_fn):
            _set_metadata_fn(d)
    else:
        raise KeyError()



from __future__ import annotations

# standard libraries
import asyncio
import copy
import gettext
import operator
import pkgutil
import threading
import typing

# third party libraries
import numpy.typing

# local libraries
from nion.data import Image
from nion.swift import MimeTypes
from nion.swift import Panel
from nion.swift import Thumbnails
from nion.swift.model import DataItem
from nion.swift.model import DisplayItem
from nion.swift.model import Persistence
from nion.swift.model import UISettings
from nion.ui import Bitmap
from nion.ui import CanvasItem
from nion.ui import DrawingContext
from nion.ui import GridCanvasItem
from nion.ui import GridFlowCanvasItem
from nion.ui import ListCanvasItem
from nion.ui import UserInterface
from nion.utils import Event
from nion.utils import Geometry
from nion.utils import Model
from nion.utils import ReferenceCounting
from nion.utils import Stream

if typing.TYPE_CHECKING:
    from nion.swift import DocumentController
    from nion.utils import Selection

_NDArray = numpy.typing.NDArray[typing.Any]

_ = gettext.gettext


"""
    The data panel has two parts:

    (1) a selection of what collection is displayed, which may be a data group, smart data group,
    or the whole document. If the whole document, then an optional filter may also be applied.
    "within the last 24 hours" would be an example filter.

    (2) the list of display items from the collection. the user may further refine the list of
    items by filtering by additional criteria. the user also chooses the sorting on the list of
    display items.

"""


class DataPanelItemBaseCanvasItem(CanvasItem.AbstractCanvasItem):
    """Canvas item to draw a data panel list item.

    This is critical performance code. It is called for every item in the list. So use a custom renderer.
    """

    def __init__(self, display_item: DisplayItem.DisplayItem, ui: UserInterface.UserInterface, font_metrics_fn: typing.Callable[[str, str], UserInterface.FontMetrics]) -> None:
        super().__init__()
        self.__display_item = display_item
        self.__ui = ui
        self.__font_metrics_fn = font_metrics_fn
        self.__thumbnail: typing.Optional[Bitmap.Bitmap] = None

        def thumbnail_updated() -> None:
            bitmap_data = self.__thumbnail_source.thumbnail_data if self.__thumbnail_source else None
            self.__thumbnail = Bitmap.Bitmap(rgba_bitmap_data=bitmap_data)
            self.update()

        self.__thumbnail_source = Thumbnails.ThumbnailManager().thumbnail_source_for_display_item(self.__ui, self.__display_item)
        self.__thumbnail_updated_event_listener = self.__thumbnail_source.thumbnail_updated_event.listen(thumbnail_updated)

        thumbnail_updated()

        self.__item_changed_listener = display_item.item_changed_event.listen(ReferenceCounting.weak_partial(DataPanelListItem.__item_changed, self))

    def _get_composer(self, composer_cache: CanvasItem.ComposerCache) -> CanvasItem.BaseComposer:
        raise NotImplementedError()

    def close(self) -> None:
        self.__thumbnail_updated_event_listener = typing.cast(typing.Any, None)
        self.__thumbnail_source = typing.cast(typing.Any, None)
        super().close()

    @property
    def _thumbnail(self) -> typing.Optional[Bitmap.Bitmap]:
        return self.__thumbnail

    @property
    def display_item(self) -> DisplayItem.DisplayItem:
        return self.__display_item

    @property
    def title(self) -> str:
        return self.__display_item.displayed_title

    @property
    def format_str(self) -> str:
        format_str = self.__display_item.size_and_data_format_as_string if self.__display_item else str()
        storage_space_string = self.__display_item.storage_space_string if self.__display_item else str()
        return " ".join([format_str, storage_space_string])

    @property
    def datetime_str(self) -> str:
        return self.__display_item.date_for_sorting_local_as_string if self.__display_item else str()

    @property
    def status_str(self) -> str:
        return self.__display_item.status_str if self.__display_item else str()

    def __item_changed(self) -> None:
        self.update()


class DataPanelListItemComposer(CanvasItem.BaseComposer):
    def __init__(self, canvas_item: CanvasItem.AbstractCanvasItem, layout_sizing: CanvasItem.Sizing, cache: CanvasItem.ComposerCache,
                 thumbnail: typing.Optional[Bitmap.Bitmap], line_height: int, displayed_title: str, format_str: str, datetime_str: str, status_str: str) -> None:
        super().__init__(canvas_item, layout_sizing, cache)
        self.__bitmap = thumbnail
        self.__line_height = line_height
        self.__displayed_title = displayed_title
        self.__format_str = format_str
        self.__datetime_str = datetime_str
        self.__status_str = status_str

    def _repaint(self, drawing_context: DrawingContext.DrawingContext, canvas_bounds: Geometry.IntRect, composer_cache: CanvasItem.ComposerCache) -> None:
        text_font = "11px sans-serif"
        text_color = "black"

        line_number = 1
        line_height = self.__line_height + 2

        if self.__displayed_title:
            drawing_context.font = text_font
            drawing_context.text_baseline = "bottom"
            drawing_context.text_align = "left"
            drawing_context.fill_style = text_color
            drawing_context.fill_text(self.__displayed_title, 82, line_height * line_number)
            line_number += 1

        if self.__format_str:
            drawing_context.font = text_font
            drawing_context.text_baseline = "bottom"
            drawing_context.text_align = "left"
            drawing_context.fill_style = text_color
            drawing_context.fill_text(self.__format_str, 82, line_height * line_number)
            line_number += 1

        if self.__datetime_str:
            drawing_context.font = text_font
            drawing_context.text_baseline = "bottom"
            drawing_context.text_align = "left"
            drawing_context.fill_style = text_color
            drawing_context.fill_text(self.__datetime_str, 82, line_height * line_number)
            line_number += 1

        if self.__status_str:
            drawing_context.font = text_font
            drawing_context.text_baseline = "bottom"
            drawing_context.text_align = "left"
            drawing_context.fill_style = text_color
            drawing_context.fill_text(self.__status_str, 82, line_height * line_number)
            line_number += 1

        if self.__bitmap and self.__bitmap.rgba_bitmap_data is not None:
            image_size = self.__bitmap.computed_shape
            if image_size.height > 0 and image_size.width > 0:
                rect = Geometry.IntRect(Geometry.IntPoint(y=4, x=4), size=Geometry.IntSize(h=72, w=72))
                display_rect = Geometry.fit_to_size(rect, image_size)
                display_height = display_rect.height
                display_width = display_rect.width
                if display_rect and display_width > 0 and display_height > 0:
                    display_top = display_rect.top
                    display_left = display_rect.left
                    drawing_context.draw_image(self.__bitmap.rgba_bitmap_data, display_left, display_top, display_width, display_height)


class DataPanelListItem(DataPanelItemBaseCanvasItem):

    def __init__(self, display_item: DisplayItem.DisplayItem, ui: UserInterface.UserInterface, font_metrics_fn: typing.Callable[[str, str], UserInterface.FontMetrics]) -> None:
        super().__init__(display_item, ui, font_metrics_fn)
        self.__font_metrics_fn = font_metrics_fn
        self.update_sizing(self.sizing.with_fixed_height(72))
        self.update_sizing(self.sizing.with_fixed_width(CanvasItem.SizingEnum.UNRESTRAINED))

    def _get_composer(self, composer_cache: CanvasItem.ComposerCache) -> CanvasItem.BaseComposer:
        line_height = self.__font_metrics_fn("11px sans-serif", "M").height
        return DataPanelListItemComposer(self, self.layout_sizing, composer_cache, self._thumbnail, line_height, self.title, self.format_str, self.datetime_str, self.status_str)


class DataPanelGridItemComposer(CanvasItem.BaseComposer):
    def __init__(self, canvas_item: CanvasItem.AbstractCanvasItem, layout_sizing: CanvasItem.Sizing,
                 cache: CanvasItem.ComposerCache, ui_settings: UISettings.UISettings,
                 thumbnail: typing.Optional[Bitmap.Bitmap], line_height: int, displayed_title: str, format_str: str,
                 datetime_str: str, status_str: str, draw_label: bool) -> None:
        super().__init__(canvas_item, layout_sizing, cache)
        self.__ui_settings = ui_settings
        self.__bitmap = thumbnail
        self.__line_height = line_height
        self.__displayed_title = displayed_title
        self.__format_str = format_str
        self.__datetime_str = datetime_str
        self.__status_str = status_str
        self.__draw_label = draw_label

    def _repaint(self, drawing_context: DrawingContext.DrawingContext, canvas_bounds: Geometry.IntRect, composer_cache: CanvasItem.ComposerCache) -> None:
        if self.__bitmap and self.__bitmap.rgba_bitmap_data is not None:
            image_size = self.__bitmap.computed_shape
            if image_size.height > 0 and image_size.width > 0:
                rect = canvas_bounds.inset(4, 4)
                if self.__draw_label:
                    rect = Geometry.IntRect.from_tlhw(rect.top, rect.left, rect.height - self.__line_height, rect.width)
                display_rect = Geometry.fit_to_size(rect, image_size)
                display_height = display_rect.height
                display_width = display_rect.width
                if display_rect and display_width > 0 and display_height > 0:
                    display_top = display_rect.top
                    display_left = display_rect.left
                    drawing_context.draw_image(self.__bitmap.rgba_bitmap_data, display_left, display_top, display_width, display_height)
                    if self.__draw_label:
                        font = "11px sans-serif"
                        drawing_context.font = font
                        drawing_context.fill_style = "black"
                        truncated_displayed_title = self.__ui_settings.truncate_string_to_width(font, self.__displayed_title, canvas_bounds.width, UISettings.TruncateModeType.MIDDLE)
                        drawing_context.text_baseline = "middle"
                        drawing_context.text_align = "center"
                        drawing_context.fill_text(truncated_displayed_title, canvas_bounds.center.x, (rect.bottom + 4 + canvas_bounds.bottom) // 2)


class DataPanelGridItem(DataPanelItemBaseCanvasItem):

    def __init__(self, display_item: DisplayItem.DisplayItem, ui: UserInterface.UserInterface, ui_settings: UISettings.UISettings, draw_label: bool = True) -> None:
        super().__init__(display_item, ui, typing.cast(typing.Callable[[str, str], UserInterface.FontMetrics], ui_settings.get_font_metrics))
        self.__ui_settings = ui_settings
        self.__draw_label = draw_label
        self.update_sizing(self.sizing.with_fixed_height(CanvasItem.SizingEnum.UNRESTRAINED))
        self.update_sizing(self.sizing.with_fixed_width(CanvasItem.SizingEnum.UNRESTRAINED))

    def _get_composer(self, composer_cache: CanvasItem.ComposerCache) -> CanvasItem.BaseComposer:
        # return CanvasItem.EmptyCanvasItemComposer(self, self.layout_sizing, composer_cache)
        line_height = self.__ui_settings.get_font_metrics("11px sans-serif", "M").height
        return DataPanelGridItemComposer(self, self.layout_sizing, composer_cache, self.__ui_settings, self._thumbnail, line_height, self.title, self.format_str, self.datetime_str, self.status_str, self.__draw_label)


class DataPanelUISettings(UISettings.UISettings):
    def __init__(self, ui: UserInterface.UserInterface) -> None:
        self.__ui = ui

    def get_font_metrics(self, font: str, text: str) -> UISettings.FontMetrics:
        return typing.cast(UISettings.FontMetrics, self.__ui.get_font_metrics(font, text))

    def truncate_string_to_width(self, font_str: str, text: str, pixel_width: int, mode: UISettings.TruncateModeType) -> str:
        return self.__ui.truncate_string_to_width(font_str, text, pixel_width, typing.cast(UserInterface.TruncateModeType, mode))

    @property
    def cursor_tolerance(self) -> float:
        return self.__ui.get_tolerance(UserInterface.ToleranceType.CURSOR)


class DataPanel(Panel.Panel):

    def __init__(self, document_controller: DocumentController.DocumentController, panel_id: str, properties: Persistence.PersistentDictType) -> None:
        super().__init__(document_controller, panel_id, _("Data Items"))

        ui = document_controller.ui

        display_items_model = document_controller.filtered_display_items_model

        self.__selection = self.document_controller.selection

        def selection_changed() -> None:
            # called when the selection changes; notify selected display item changed if focused.
            self.__notify_focus_changed()

        self.__selection_changed_event_listener = self.__selection.changed_event.listen(selection_changed)

        class ItemDelegate(GridFlowCanvasItem.GridFlowCanvasItemDelegate):
            def __init__(self, data_panel: DataPanel, selection: Selection.IndexedSelection) -> None:
                self.__data_panel = data_panel
                self.__selection = selection

            def item_tool_tip(self, item: typing.Any) -> typing.Optional[str]:
                display_item = typing.cast(DisplayItem.DisplayItem, item)
                return display_item.list_tool_tip_str

            def context_menu_event(self, context_menu_event: GridFlowCanvasItem.GridFlowCanvasItemContextMenuEvent) -> bool:
                display_items = tuple(typing.cast(DisplayItem.DisplayItem, item) for item in context_menu_event.selected_items)
                menu = document_controller.create_context_menu()
                action_context = document_controller._get_action_context_for_display_items(display_items, None)
                document_controller.populate_context_menu(menu, action_context)
                menu.add_separator()
                document_controller.add_action_to_menu(menu, "item.delete", action_context)
                menu.popup(context_menu_event.gp.x, context_menu_event.gp.y)
                return True

            def delete_event(self, delete_event: GridFlowCanvasItem.GridFlowCanvasItemDeleteEvent) -> bool:
                display_items = tuple(typing.cast(DisplayItem.DisplayItem, item) for item in delete_event.selected_items)
                document_controller.delete_display_items(display_items)
                return True

            def drag_started_event(self, drag_started_event: GridFlowCanvasItem.GridFlowCanvasItemDragStartedEvent) -> bool:
                mime_data, thumbnail_data = self._get_mime_data_and_thumbnail_data(drag_started_event)
                if mime_data:
                    self.__data_panel.widget.drag(mime_data, thumbnail_data)
                    return True
                return False

            def can_drop_mime_data(self, mime_data: UserInterface.MimeData, action: str, drop_index: int | None) -> bool:
                return mime_data.has_file_paths

            def drop_mime_data(self, mime_data: UserInterface.MimeData, action: str, drop_index: int | None) -> str:
                display_items = document_controller.receive_files(mime_data.file_paths)
                if set(display_items).intersection(set(document_controller.filtered_display_items_model.display_items)):
                    document_controller.select_display_items_in_data_panel(display_items)
                return "accept"

            def _get_mime_data_and_thumbnail_data(self, drag_started_event: GridFlowCanvasItem.GridFlowCanvasItemDragStartedEvent) -> typing.Tuple[typing.Optional[UserInterface.MimeData], typing.Optional[_NDArray]]:
                mime_data = None
                thumbnail_data = None
                display_item = typing.cast(DisplayItem.DisplayItem, drag_started_event.item)
                display_items = tuple(
                    typing.cast(DisplayItem.DisplayItem, item) for item in drag_started_event.selected_items)
                if len(display_items) <= 1 and display_item is not None:
                    mime_data = document_controller.ui.create_mime_data()
                    MimeTypes.mime_data_put_display_item(mime_data, display_item)
                    thumbnail_source = Thumbnails.ThumbnailManager().thumbnail_source_for_display_item(self.__data_panel.ui, display_item)
                    thumbnail_data = thumbnail_source.thumbnail_data
                    if thumbnail_data is not None:
                        # scaling is very slow
                        thumbnail_data = Image.get_rgba_data_from_rgba(
                            Image.scaled(Image.get_rgba_view_from_rgba_data(thumbnail_data),
                                         tuple(Geometry.IntSize(w=80, h=80))))
                elif len(display_items) > 1:
                    mime_data = document_controller.ui.create_mime_data()
                    MimeTypes.mime_data_put_display_items(mime_data, display_items)
                    anchor_index = self.__selection.anchor_index or 0
                    thumbnail_display_item = display_items[anchor_index]
                    thumbnail_source = Thumbnails.ThumbnailManager().thumbnail_source_for_display_item(self.__data_panel.ui, thumbnail_display_item)
                    thumbnail_data = thumbnail_source.thumbnail_data
                return mime_data, thumbnail_data

        item_delegate = ItemDelegate(self, self.__selection)

        def list_item_factory(item: typing.Any, is_selected_model: Model.PropertyModel[bool]) -> CanvasItem.AbstractCanvasItem:
            return DataPanelListItem(typing.cast(DisplayItem.DisplayItem, item), document_controller.ui, document_controller.get_font_metrics)

        # note is_shared_selection is True for both list and grid canvas items. prevents the selection from being updated when items are inserted.
        # instead, the selection in the model itself is used.
        list_canvas_item = ListCanvasItem.ListCanvasItem2(Panel.ThreadSafeListModel(display_items_model, document_controller.event_loop), self.__selection, list_item_factory, item_delegate, item_height=80, key="display_items", is_shared_selection=True)
        list_canvas_item.wants_drag_events = True
        list_scroll_area_canvas_item = CanvasItem.ScrollAreaCanvasItem(list_canvas_item)
        list_scroll_bar_canvas_item = CanvasItem.ScrollBarCanvasItem(list_scroll_area_canvas_item, CanvasItem.Orientation.Vertical)
        list_scroll_group_canvas_item = CanvasItem.CanvasItemComposition()
        list_scroll_group_canvas_item.layout = CanvasItem.CanvasItemRowLayout()
        list_scroll_group_canvas_item.add_canvas_item(list_scroll_area_canvas_item)
        list_scroll_group_canvas_item.add_canvas_item(list_scroll_bar_canvas_item)

        def grid_item_factory(item: typing.Any, is_selected_model: Model.PropertyModel[bool]) -> CanvasItem.AbstractCanvasItem:
            return DataPanelGridItem(typing.cast(DisplayItem.DisplayItem, item), document_controller.ui, DataPanelUISettings(document_controller.ui))

        # note is_shared_selection is True for both list and grid canvas items. prevents the selection from being updated when items are inserted.
        # instead, the selection in the model itself is used.
        line_height = document_controller.get_font_metrics("11px sans-serif", "M").height
        grid_canvas_item = GridCanvasItem.GridCanvasItem2(Panel.ThreadSafeListModel(display_items_model, document_controller.event_loop), self.__selection, grid_item_factory, item_delegate, item_size=Geometry.IntSize(80 + line_height, 80), key="display_items", is_shared_selection=True)
        grid_canvas_item.wants_drag_events = True
        grid_scroll_area_canvas_item = CanvasItem.ScrollAreaCanvasItem(grid_canvas_item)
        grid_scroll_area_canvas_item.auto_resize_contents = True
        grid_scroll_bar_canvas_item = CanvasItem.ScrollBarCanvasItem(grid_scroll_area_canvas_item, CanvasItem.Orientation.Vertical)
        grid_scroll_group_canvas_item = CanvasItem.CanvasItemComposition()
        grid_scroll_group_canvas_item.layout = CanvasItem.CanvasItemRowLayout()
        grid_scroll_group_canvas_item.add_canvas_item(grid_scroll_area_canvas_item)
        grid_scroll_group_canvas_item.add_canvas_item(grid_scroll_bar_canvas_item)

        def begin_changes(key: str) -> None:
            list_canvas_item._begin_batch_update()
            grid_canvas_item._begin_batch_update()

        def end_changes(key: str) -> None:
            list_canvas_item._end_batch_update()
            grid_canvas_item._end_batch_update()

        # the display items model can notify us when it is about to change. in order to gang up changes, watch
        # for these notification events and tell the list canvas item to only update at the end of the changes.
        self.__display_items_begin_changes_listener = display_items_model.begin_changes_event.listen(begin_changes)
        self.__display_items_end_changes_listener = display_items_model.end_changes_event.listen(end_changes)

        # for testing
        self._scroll_area_canvas_item = list_scroll_area_canvas_item
        self._scroll_bar_canvas_item = list_scroll_bar_canvas_item

        list_icon_20_bytes = pkgutil.get_data(__name__, "resources/list_icon_20.png")
        grid_icon_20_bytes = pkgutil.get_data(__name__, "resources/grid_icon_20.png")
        list_icon_20_bytes = list_icon_20_bytes or bytes()
        grid_icon_20_bytes = grid_icon_20_bytes or bytes()
        list_icon_button = CanvasItem.BitmapButtonCanvasItem(CanvasItem.load_rgba_data_from_bytes(list_icon_20_bytes))
        grid_icon_button = CanvasItem.BitmapButtonCanvasItem(CanvasItem.load_rgba_data_from_bytes(grid_icon_20_bytes))

        list_icon_button.update_sizing(list_icon_button.sizing.with_fixed_size(Geometry.IntSize(20, 20)))
        grid_icon_button.update_sizing(grid_icon_button.sizing.with_fixed_size(Geometry.IntSize(20, 20)))

        button_row = CanvasItem.CanvasItemComposition()
        button_row.layout = CanvasItem.CanvasItemRowLayout(spacing=4)
        button_row.add_canvas_item(list_icon_button)
        button_row.add_canvas_item(grid_icon_button)

        buttons_widget = ui.create_canvas_widget(properties={"height": 20, "width": 44})
        buttons_widget.canvas_item.add_canvas_item(button_row)

        search_widget = ui.create_row_widget()
        search_widget.add_spacing(8)
        search_widget.add(ui.create_label_widget(_("Filter")))
        search_widget.add_spacing(8)
        search_line_edit = ui.create_line_edit_widget()
        search_line_edit.placeholder_text = _("No Filter")
        search_line_edit.clear_button_enabled = True  # Qt 5.3 doesn't signal text edited or editing finished when clearing. useless so disabled.
        search_line_edit.on_text_edited = self.document_controller.filter_controller.text_filter_changed
        search_line_edit.on_editing_finished = self.document_controller.filter_controller.text_filter_changed
        search_widget.add(search_line_edit)
        search_widget.add_spacing(6)
        search_widget.add(buttons_widget)
        search_widget.add_spacing(8)

        stack_canvas_item = CanvasItem.StackCanvasItem()
        stack_canvas_item.add_canvas_item(list_scroll_group_canvas_item)
        stack_canvas_item.add_canvas_item(grid_scroll_group_canvas_item)
        stack_canvas_item.current_index = 0
        stack_canvas_item.update_sizing(stack_canvas_item.sizing.with_unconstrained_height())

        self.data_view_widget = ui.create_canvas_widget(properties={"size-policy-vertical": "expanding"})
        self.data_view_widget.canvas_item.add_canvas_item(stack_canvas_item)

        self.__view_button_group = CanvasItem.RadioButtonGroup([list_icon_button, grid_icon_button])
        self.__view_button_group.current_index = 0
        self.__view_button_group.on_current_index_changed = lambda index: setattr(stack_canvas_item, "current_index", index)

        self.__filter_description_combo_box = ui.create_combo_box_widget(item_getter=operator.attrgetter("title"))

        status_row = ui.create_row_widget()
        status_row.add_spacing(4)
        status_row.add(self.__filter_description_combo_box)
        status_row.add_spacing(3)

        divider_widget = ui.create_canvas_widget(properties={"height": 2})
        divider_widget.canvas_item.add_canvas_item(CanvasItem.DividerCanvasItem(orientation="horizontal"))

        self.__status_section = ui.create_column_widget()
        self.__status_section.add_spacing(2)
        self.__status_section.add(status_row)
        self.__status_section.add_spacing(4)
        self.__status_section.add(divider_widget)

        widget = ui.create_column_widget(properties=properties)
        widget.add(self.__status_section)
        widget.add(self.data_view_widget)
        widget.add_spacing(6)
        widget.add(search_widget)
        widget.add_spacing(6)

        self.widget = widget

        self.__list_canvas_item = list_canvas_item
        self.__grid_canvas_item = grid_canvas_item

        # listen to the focus changed event for the list and grid canvas items.
        # if we are receiving focus, tell the window (document_controller) to update the selected display item.
        self.__list_canvas_item_focus_changed_event_listener = list_canvas_item.focus_changed_event.listen(self.__notify_focus_changed)
        self.__grid_canvas_item_focus_changed_event_listener = grid_canvas_item.focus_changed_event.listen(self.__notify_focus_changed)

        # for tests only
        self._data_list_canvas_item = list_scroll_group_canvas_item
        self._data_grid_canvas_item = grid_scroll_group_canvas_item

        def update_filter_description(collection_info: typing.Optional[DocumentController.CollectionInfo]) -> None:
            if collection_info != self.__filter_description_combo_box.current_item:
                self.__filter_description_combo_box.current_item = collection_info

        def update_filter_descriptions(collection_info_list: typing.Optional[typing.Sequence[DocumentController.CollectionInfo]]) -> None:
            if collection_info_list is not None:
                self.__filter_description_combo_box.items = collection_info_list
                update_filter_description(document_controller.current_collection_info.value)

        def on_current_collection_changed(collection_info: typing.Optional[DocumentController.CollectionInfo]) -> None:
            if collection_info and document_controller.current_collection_info.value != collection_info:
                if collection_info.is_smart_collection:
                    document_controller.set_filter(collection_info.filter_id)
                else:
                    document_controller.set_data_group(collection_info.data_group)

        self.__filter_description_action = Stream.ValueStreamAction(document_controller.current_collection_info, update_filter_description)
        collection_info_list_stream = Stream.PropertyChangedEventStream[typing.Sequence["DocumentController.CollectionInfo"]](document_controller.collection_info_list_model, "items")
        self.__collection_info_list_stream_action = Stream.ValueStreamAction(collection_info_list_stream, update_filter_descriptions)

        update_filter_descriptions(document_controller.collection_info_list_model.items)
        self.__filter_description_combo_box.on_current_item_changed = on_current_collection_changed

    def close(self) -> None:
        self.__selection_changed_event_listener.close()
        self.__selection_changed_event_listener = typing.cast(Event.EventListener, None)
        self.__filter_description_action = typing.cast(typing.Any, None)
        self.__view_button_group.close()
        self.__view_button_group = typing.cast(CanvasItem.RadioButtonGroup, None)
        # close the widget to stop repainting the widgets before closing the controllers.
        super().close()

    @property
    def _selection(self) -> Selection.IndexedSelection:
        return self.__selection

    @property
    def _list_canvas_item(self) -> ListCanvasItem.ListCanvasItem2:
        return self.__list_canvas_item

    @property
    def _grid_canvas_item(self) -> GridCanvasItem.GridCanvasItem2:
        return self.__grid_canvas_item

    def __notify_focus_changed(self) -> None:
        # this is called when the keyboard focus for the data panel is changed.
        # if we are receiving focus, tell the window (document_controller) that
        # we now have the focus.
        if self.__list_canvas_item.focused or self.__grid_canvas_item.focused:
            self.document_controller.data_panel_focused()

    def _request_focus_for_test(self) -> None:
        self.__list_canvas_item.request_focus()

    def make_selection_visible(self) -> None:
        self.__list_canvas_item.make_selection_visible()
        self.__grid_canvas_item.make_selection_visible()



from __future__ import annotations

# system imports
import code
import contextlib
import copy
import dataclasses
import gettext
import importlib
import io
import logging
import re
import sys
import types
import typing

# local libraries
from nion.swift import Panel
from nion.swift.model import DocumentModel
from nion.swift.model import Persistence
from nion.ui import Dialog
from nion.ui import UserInterface
from nion.ui import Widgets
from nion.utils import Registry

# hack to work with conda/python3.10 until they fix readline
rlcompleter: typing.Optional[types.ModuleType]
try:
    rlcompleter = importlib.import_module("rlcompleter")
except Exception:
    rlcompleter = None

if typing.TYPE_CHECKING:
    from nion.swift import DocumentController

_ = gettext.gettext


@dataclasses.dataclass
class ConsoleStartupInfo:
    console_startup_id: str
    console_startup_lines: typing.Sequence[str]
    console_startup_help: typing.Optional[typing.Sequence[str]]


class ConsoleStartupComponent(typing.Protocol):
    def get_console_startup_info(self, logger: logging.Logger) -> ConsoleStartupInfo: ...


class ConsoleWidgetStateController:
    delims = " \t\n`~!@#$%^&*()-=+[{]}\\|;:\'\",<>/?"

    def __init__(self, locals: typing.Dict[str, typing.Any]) -> None:
        self.__incomplete = False

        self.__console = code.InteractiveConsole(locals)

        self.__history: typing.List[str] = list()
        self.__history_point: typing.Optional[int] = None
        self.__command_cache: typing.Tuple[typing.Optional[int], str] = (None, str()) # Meaning of the tuple: (history_point where the command belongs, command)

    def close(self) -> None:
        # through experimentation, this is how to ensure the locals are
        # garbage collected upon closing.
        typing.cast(typing.Dict[str, typing.Any], self.__console.locals).clear()
        self.__console.locals = typing.cast(typing.Any, None)
        self.__console = typing.cast(typing.Any, None)

    @staticmethod
    def get_common_prefix(l: typing.List[str]) -> str:
        if not l:
            return str()
        s1 = min(l)  # return the first, alphabetically
        s2 = max(l)  # the last
        # check common characters in between
        for i, c in enumerate(s1):
            if c != s2[i]:
                return s1[:i]
        return s1

    @property
    def incomplete(self) -> bool:
        return self.__incomplete

    # interpretCommand is called from the intrinsic widget.
    def interpret_command(self, command: str) -> typing.Tuple[str, int]:
        if command:
            self.__history.append(command)
        self.__history_point = None
        self.__command_cache = (None, str())
        output = io.StringIO()
        error = io.StringIO()
        with contextlib.redirect_stdout(output), contextlib.redirect_stderr(error):
            self.__incomplete = self.__console.push(command)
        if error.getvalue():
            result =  error.getvalue()
            error_code = -1
        else:
            result = output.getvalue()
            error_code = 0
        return result, error_code

    def complete_command(self, command: str) -> typing.Tuple[str, typing.List[str]]:
        terms = list()
        completed_command = command
        if rlcompleter:
            completer = rlcompleter.Completer(namespace=getattr(self.__console, "locals"))  # TODO: why isn't locals defined?
            index = 0
            rx = "([" + re.escape(ConsoleWidgetStateController.delims) + "])"
            # the parenthesis around rx make it a group. This will cause split to keep the characters in rx in the
            # list, so that we can reconstruct the original string later
            split_commands = re.split(rx, command)
            if len(split_commands) > 0:
                completion_term = split_commands[-1]
                while True:
                    term = completer.complete(completion_term, index)
                    if term is None:
                        break
                    index += 1
                    # for some reason rlcomplete returns "\t" when completing "", so exclude that case here
                    if not term.startswith(completion_term + "__") and term != "\t":
                        terms.append(term)
                if len(terms) == 1:
                    completed_command = command[:command.rfind(completion_term)] + terms[0]
                    terms = list()
                elif len(terms) > 1:
                    common_prefix = ConsoleWidgetStateController.get_common_prefix(terms)
                    completed_command = str().join(split_commands[:-1]) + common_prefix
        return completed_command, terms

    def move_back_in_history(self, current_line: str) -> str:
        line = str()
        if self.__history_point is None:
            self.__history_point = len(self.__history)
            # do not update command_cache if the user didn't type anything
            if current_line:
                self.__command_cache = (None, current_line)
        elif self.__history_point < len(self.__history):
            # This means the user changed something at the current point in history. Save the temporary command.
            if current_line != self.__history[self.__history_point]:
                self.__command_cache = (self.__history_point, current_line)
        self.__history_point = max(0, self.__history_point - 1)
        if self.__history_point < len(self.__history):
            line = self.__command_cache[1] if self.__command_cache[0] == self.__history_point else self.__history[self.__history_point]
        return line

    def move_forward_in_history(self, current_line: str) -> str:
        line = str()
        if self.__history_point is not None:
            if self.__history_point < len(self.__history):
                # This means the user changed something at the current point in history.
                # Save the temporary command, but only if the user actually typed something
                if current_line and current_line != self.__history[self.__history_point]:
                    self.__command_cache = (self.__history_point, current_line)
            self.__history_point = min(len(self.__history), self.__history_point + 1)
            if self.__history_point < len(self.__history):
                line = self.__command_cache[1] if self.__command_cache[0] == self.__history_point else self.__history[self.__history_point]
            else:
                self.__history_point = None
        # Do not use 'else' here because history_point might have been set to 'None' in the first 'if' statement
        if self.__history_point is None:
            line = self.__command_cache[1] if self.__command_cache[0] is None else str()
        return line


class ConsoleWidget(Widgets.CompositeWidgetBase):

    def __init__(self,
                 ui: UserInterface.UserInterface,
                 logger: logging.Logger,
                 locals: typing.Optional[typing.Dict[str, typing.Any]] = None,
                 properties: typing.Optional[Persistence.PersistentDictType] = None) -> None:
        content_widget = ui.create_column_widget()
        super().__init__(content_widget)

        self.__logger = logger
        self.__closed = False

        self.prompt = ">>> "
        self.continuation_prompt = "... "

        self.__cursor_position: typing.Optional[UserInterface.CursorPosition] = None

        self.__text_edit_widget = ui.create_text_edit_widget(properties)
        self.__text_edit_widget.set_text_color("white")
        self.__text_edit_widget.set_text_background_color("black")
        self.__text_edit_widget.set_text_font(Panel.Panel.get_monospace_text_font())
        self.__text_edit_widget.set_line_height_proportional(Panel.Panel.get_monospace_proportional_line_height())
        self.__text_edit_widget.word_wrap_mode = "anywhere"
        self.__text_edit_widget.on_cursor_position_changed = self.__cursor_position_changed
        self.__text_edit_widget.on_return_pressed = self.__return_pressed
        self.__text_edit_widget.on_key_pressed = self.__key_pressed
        self.__text_edit_widget.on_insert_mime_data = self.__insert_mime_data

        class StdoutCatcher:
            def __init__(self, text_edit_widget: UserInterface.TextEditWidget) -> None:
                self.__text_edit_widget = text_edit_widget

            def write(self, stuff: str) -> int:
                if self.__text_edit_widget:
                    stripped_stuff = str(stuff).rstrip()
                    self.__text_edit_widget.append_text(stripped_stuff)
                    return len(stripped_stuff)
                return 0

            def flush(self) -> None:
                pass

        stdout = StdoutCatcher(self.__text_edit_widget)

        class LoggingHandler(logging.Handler):

            def __init__(self, console_widget: ConsoleWidget) -> None:
                super().__init__()
                self.__console_widget = console_widget
                self._prompted = False

            def emit(self, record: logging.LogRecord) -> None:
                if record.levelno >= logging.INFO:
                    color = None if self._prompted else "grey"
                    self.__console_widget.append_line(str(record.getMessage()).rstrip(), color=color)

        self.__logging_handler = LoggingHandler(self)
        self.__logger.addHandler(self.__logging_handler)

        locals = locals if locals is not None else dict()
        locals.update({'__name__': None, '__console__': None, '__doc__': None, '_stdout': stdout})

        self.__state_controller = ConsoleWidgetStateController(locals)

        self.__last_cursor_position: typing.Optional[UserInterface.CursorPosition] = None

        content_widget.add(self.__text_edit_widget)

    def close(self) -> None:
        self.__closed = True
        # be sure to clear callbacks in the text widget which may use the text edit widget.
        self.__text_edit_widget.on_cursor_position_changed = None
        self.__text_edit_widget.on_return_pressed = None
        self.__text_edit_widget.on_key_pressed = None
        self.__text_edit_widget.on_insert_mime_data = None
        super().close()
        self.__text_edit_widget = typing.cast(typing.Any, None)
        self.__state_controller.close()
        self.__state_controller = typing.cast(typing.Any, None)

    def show_prompt(self) -> None:
        self.__logging_handler._prompted = True
        self.__text_edit_widget.append_text(self.prompt)
        self.__text_edit_widget.move_cursor_position("end")
        self.__last_cursor_position = copy.deepcopy(self.__cursor_position)

    def activate(self) -> None:
        self.__text_edit_widget.focused = True

    @property
    def current_prompt(self) -> str:
        return self.continuation_prompt if self.__state_controller.incomplete else self.prompt

    def insert_lines(self, lines: typing.Sequence[str]) -> None:
        for l in lines:
            text_edit_widget = self.__text_edit_widget
            text_edit_widget.move_cursor_position("end")
            text_edit_widget.insert_text(l)
            result, error_code = self.__state_controller.interpret_command(l)
            # if the console is closed as a result of running the script, don't try to append text.
            # the console can be closed as a result of switching to another project via script.
            if not self.__closed:
                if len(result) > 0:
                    text_edit_widget.set_text_color("red" if error_code else "aquamarine")
                    text_edit_widget.append_text(result[:-1])
                    text_edit_widget.set_text_color("white")
                text_edit_widget.append_text(self.current_prompt)
                text_edit_widget.move_cursor_position("end")
                self.__last_cursor_position = copy.deepcopy(self.__cursor_position)

    def append_line(self, line: str, *, color: typing.Optional[str] = None) -> None:
        self.__text_edit_widget.set_text_color(color if color else "aquamarine")
        self.__text_edit_widget.append_text(line)
        self.__text_edit_widget.set_text_color("white")

    def interpret_lines(self, lines: typing.Sequence[str]) -> None:
        for l in lines:
            self.__state_controller.interpret_command(l)

    def __return_pressed(self) -> bool:
        command = self.__get_partial_command()
        result, error_code = self.__state_controller.interpret_command(command)
        if len(result) > 0:
            self.__text_edit_widget.set_text_color("red" if error_code else "aquamarine")
            self.__text_edit_widget.append_text(result[:-1])
        self.__text_edit_widget.set_text_color("white")
        self.__text_edit_widget.append_text(self.current_prompt)
        self.__text_edit_widget.move_cursor_position("end")
        self.__last_cursor_position = copy.deepcopy(self.__cursor_position)
        return True

    def __get_partial_command(self) -> str:
        text = self.__text_edit_widget.text
        command = text.split('\n')[-1] if text else str()
        if command.startswith(self.prompt):
            command = command[len(self.prompt):]
        elif command.startswith(self.continuation_prompt):
            command = command[len(self.continuation_prompt):]
        return command

    def __key_pressed(self, key: UserInterface.Key) -> bool:
        block_number = self.__cursor_position.block_number if self.__cursor_position else None
        last_block_number = self.__last_cursor_position.block_number if self.__last_cursor_position else None
        is_cursor_on_last_line = block_number == last_block_number
        partial_command = self.__get_partial_command()
        column_number = self.__cursor_position.column_number if self.__cursor_position else 0
        is_cursor_on_last_column = (partial_command.strip() and column_number == len(self.current_prompt + partial_command))

        if is_cursor_on_last_line and key.is_up_arrow:
            line = self.__state_controller.move_back_in_history(partial_command)
            self.__text_edit_widget.move_cursor_position("start_para", "move")
            self.__text_edit_widget.move_cursor_position("end_para", "keep")
            self.__text_edit_widget.insert_text("{}{}".format(self.current_prompt, line))
            self.__text_edit_widget.move_cursor_position("end")
            self.__last_cursor_position = copy.deepcopy(self.__cursor_position)
            return True

        if is_cursor_on_last_line and key.is_down_arrow:
            line = self.__state_controller.move_forward_in_history(partial_command)
            self.__text_edit_widget.move_cursor_position("start_para", "move")
            self.__text_edit_widget.move_cursor_position("end_para", "keep")
            self.__text_edit_widget.insert_text("{}{}".format(self.current_prompt, line))
            self.__text_edit_widget.move_cursor_position("end")
            self.__last_cursor_position = copy.deepcopy(self.__cursor_position)
            return True

        if is_cursor_on_last_line and key.is_delete:
            if not partial_command:
                return True

        if is_cursor_on_last_line and key.is_move_to_start_of_line:
            mode = "keep" if key.modifiers.shift else "move"
            self.__text_edit_widget.move_cursor_position("start_para", mode)
            self.__text_edit_widget.move_cursor_position("next", mode, n=4)
            return True

        if is_cursor_on_last_line and key.is_delete_to_end_of_line:
            self.__text_edit_widget.move_cursor_position("end_para", "keep")
            self.__text_edit_widget.remove_selected_text()
            return True

        if is_cursor_on_last_line and key.key == 0x43 and key.modifiers.native_control and sys.platform == "darwin":
            self.__text_edit_widget.move_cursor_position("end")
            self.__text_edit_widget.insert_text("\n")
            self.__text_edit_widget.insert_text(self.current_prompt)
            self.__text_edit_widget.move_cursor_position("end")
            self.__last_cursor_position = copy.deepcopy(self.__cursor_position)
            return True

        if is_cursor_on_last_line and is_cursor_on_last_column and key.is_tab:
            completed_command, terms = self.__state_controller.complete_command(partial_command)
            if not terms:
                self.__text_edit_widget.move_cursor_position("start_para", "move")
                self.__text_edit_widget.move_cursor_position("end_para", "keep")
                self.__text_edit_widget.insert_text("{}{}".format(self.current_prompt, completed_command))
                self.__text_edit_widget.move_cursor_position("end")
                self.__last_cursor_position = copy.deepcopy(self.__cursor_position)
            elif len(terms) > 1:
                self.__text_edit_widget.move_cursor_position("end")
                self.__text_edit_widget.set_text_color("brown")
                self.__text_edit_widget.append_text("   ".join(terms) + "\n")
                self.__text_edit_widget.move_cursor_position("end")
                self.__text_edit_widget.set_text_color("white")
                self.__text_edit_widget.insert_text("{}{}".format(self.current_prompt, completed_command))
                self.__text_edit_widget.move_cursor_position("end")
                self.__last_cursor_position = copy.deepcopy(self.__cursor_position)
            return True
        return False

    def __cursor_position_changed(self, cursor_position: UserInterface.CursorPosition) -> None:
        self.__cursor_position = copy.deepcopy(cursor_position)

    def __insert_mime_data(self, mime_data: UserInterface.MimeData) -> None:
        text = mime_data.data_as_string("text/plain")
        text_lines = re.split("[" + re.escape("\n") + re.escape("\r") + "]", text)
        if not text_lines[-1]:
            text_lines = text_lines[:-1]
        if len(text_lines) == 1 and text_lines[0] == text.rstrip():
            # special case where text has no line terminator
            self.__text_edit_widget.insert_text(text)
        else:
            self.insert_lines(text_lines)


class DefaultConsoleStartupComponent:
    def get_console_startup_info(self, logger: logging.Logger) -> ConsoleStartupInfo:
        logger.info("Console Startup (logger, api, ui, show, etc.)")
        lines = [
            "import logging",
            "import numpy as np",
            "import numpy as numpy",
            "import uuid",
            "from nion.utils import Registry",
            "from nion.swift.model import PlugInManager",
            "from nion.ui import Declarative",
            "from nion.data import xdata_1_0 as xd",
            "get_api = PlugInManager.api_broker_fn",
            "api = get_api('~1.0', '~1.0')",
            "ui = Declarative.DeclarativeUI()",
            "_application = api.application._application",
            "_model = api.library._document_model",
            "_window = api.application.document_windows[0]._document_controller",
            "show = api.show",
            "def run_script(*args, **kwargs):",
            "  api.run_script(*args, stdout=_stdout, **kwargs)",
            "",
            f"logger = logging.getLogger('console-loggers-{ConsoleDialog.console_number}')",
            ]

        variable_to_item_map = DocumentModel.MappedItemManager().item_map
        for variable_name, data_item in variable_to_item_map.items():
            data_item_specifier = data_item.item_specifier
            lines.append(f"{variable_name} = api.library.get_item_by_specifier(api.create_specifier(item_uuid=uuid.UUID('{str(data_item_specifier.item_uuid)}')))")

        return ConsoleStartupInfo("default_startup", lines, None)


Registry.register_component(DefaultConsoleStartupComponent(), {"console-startup"})


class ConsoleDialog(Dialog.ActionDialog):

    console_number = 0

    def __init__(self, document_controller: DocumentController.DocumentController) -> None:
        super().__init__(document_controller.ui, _("Python Console"), parent_window=document_controller, persistent_id="ConsoleDialog")

        self.__document_controller = document_controller

        ConsoleDialog.console_number += 1
        logger = logging.getLogger(f"console-loggers-{ConsoleDialog.console_number}")

        self.__console_widget = ConsoleWidget(document_controller.ui, logger, properties={"min-height": 180, "min-width": 540})

        lines: list[str] = list()
        for component in Registry.get_components_by_type("console-startup"):
            console_startup_component = typing.cast(ConsoleStartupComponent, component)
            lines.extend(console_startup_component.get_console_startup_info(logger).console_startup_lines)

        self.__console_widget.interpret_lines(lines)

        self.__console_widget.show_prompt()

        self.content.add(self.__console_widget)

        self.__document_controller.register_console(self)

        self._create_menus()

    def close(self) -> None:
        self.__document_controller.unregister_console(self)
        super().close()

    def about_to_show(self) -> None:
        super().about_to_show()
        def request_focus() -> None:
            self.activate()
            self.__console_widget.activate()
        self.queue_task(request_focus)

    def assign_item_var(self, item_var: str, item: Persistence.PersistentObject) -> None:
        self.__console_widget.insert_lines([f"{item_var} = api.library.get_item_by_specifier(api.create_specifier(uuid.UUID(\"{item.uuid}\")))"])



from __future__ import annotations

# standard libraries
import asyncio
import typing

# third party libraries
import numpy.typing

# local libraries
from nion.data import Image
from nion.swift import MimeTypes
from nion.swift import Thumbnails
from nion.swift.model import DisplayItem
from nion.swift.model import DocumentModel
from nion.ui import CanvasItem
from nion.ui import UserInterface
from nion.ui import Widgets
from nion.ui import CanvasItem
from nion.utils import Geometry
from nion.utils import Process
from nion.utils import ReferenceCounting

if typing.TYPE_CHECKING:
    from nion.swift import DocumentController
    from nion.swift.model import Persistence
    from nion.ui import DrawingContext
    from nion.ui import Window
    from nion.utils import Binding
    from nion.utils import Event

_ImageDataType = Image._ImageDataType
_NDArray = numpy.typing.NDArray[typing.Any]


class AbstractThumbnailSource:

    def __init__(self) -> None:
        self.on_thumbnail_data_changed: typing.Optional[typing.Callable[[typing.Optional[_NDArray]], None]] = None
        self.__thumbnail_data: typing.Optional[_NDArray] = None
        self.overlay_canvas_items = list[CanvasItem.AbstractCanvasItem]()

    def close(self) -> None:
        self.on_thumbnail_data_changed = None

    @property
    def thumbnail_data(self) -> typing.Optional[_NDArray]:
        return self.__thumbnail_data

    def _set_thumbnail_data(self, thumbnail_data: typing.Optional[_NDArray]) -> None:
        self.__thumbnail_data = thumbnail_data

    def populate_mime_data_for_drag(self, mime_data: UserInterface.MimeData, size: Geometry.IntSize) -> typing.Tuple[bool, typing.Optional[_NDArray]]:
        return False, None


class BitmapOverlayCanvasItemComposer(CanvasItem.BaseComposer):
    def __init__(self, canvas_item: CanvasItem.AbstractCanvasItem, layout_sizing: CanvasItem.Sizing, cache: CanvasItem.ComposerCache, is_active: bool, is_dropping: bool, is_focused: bool) -> None:
        super().__init__(canvas_item, layout_sizing, cache)
        self.__is_active = is_active
        self.__is_dropping = is_dropping
        self.__is_focused = is_focused

    def _repaint(self, drawing_context: DrawingContext.DrawingContext, canvas_bounds: Geometry.IntRect, composer_cache: CanvasItem.ComposerCache) -> None:
        is_active = self.__is_active
        is_dropping = self.__is_dropping
        is_focused = self.__is_focused
        focused_style = "#3876D6"  # TODO: platform dependent
        with drawing_context.saver():
            drawing_context.translate(canvas_bounds.left, canvas_bounds.top)
            if is_active:
                with drawing_context.saver():
                    drawing_context.begin_path()
                    drawing_context.round_rect(2, 2, 6, 6, 3)
                    drawing_context.fill_style = "rgba(0, 255, 0, 0.80)"
                    drawing_context.fill()
            if is_dropping:
                with drawing_context.saver():
                    drawing_context.begin_path()
                    drawing_context.rect(0, 0, canvas_bounds.width, canvas_bounds.height)
                    drawing_context.fill_style = "rgba(255, 0, 0, 0.10)"
                    drawing_context.fill()
            if is_focused:
                stroke_style = focused_style
                drawing_context.begin_path()
                drawing_context.rect(2, 2, canvas_bounds.width - 4, canvas_bounds.height - 4)
                drawing_context.line_join = "miter"
                drawing_context.stroke_style = stroke_style
                drawing_context.line_width = 4.0
                drawing_context.stroke()


class BitmapOverlayCanvasItem(CanvasItem.AbstractCanvasItem):
    def __init__(self) -> None:
        super().__init__()
        self.__is_active = False
        self.__is_focused = False
        self.__is_dropping = False

    @property
    def is_active(self) -> bool:
        return self.__is_active

    @is_active.setter
    def is_active(self, value: bool) -> None:
        if value != self.__is_active:
            self.__is_active = value
            self.update()

    @property
    def is_focused(self) -> bool:
        return self.__is_focused

    @is_focused.setter
    def is_focused(self, value: bool) -> None:
        if value != self.__is_focused:
            self.__is_focused = value
            self.update()

    @property
    def is_dropping(self) -> bool:
        return self.__is_dropping

    @is_dropping.setter
    def is_dropping(self, value: bool) -> None:
        if value != self.__is_dropping:
            self.__is_dropping = value
            self.update()

    def _get_composer(self, composer_cache: CanvasItem.ComposerCache) -> typing.Optional[CanvasItem.BaseComposer]:
        return BitmapOverlayCanvasItemComposer(self, self.layout_sizing, composer_cache, self.__is_active, self.__is_dropping, self.__is_focused)


class BitmapOverlayCanvasItemComposition(CanvasItem.CanvasItemComposition):

    def __init__(self) -> None:
        super().__init__()
        self.focusable = True
        self.wants_drag_events = True
        self.wants_mouse_events = True
        self.__drag_start: typing.Optional[Geometry.IntPoint] = None
        self.on_drop_mime_data: typing.Optional[typing.Callable[[UserInterface.MimeData, int, int], str]] = None
        self.on_delete: typing.Optional[typing.Callable[[], None]] = None
        self.on_drag_pressed: typing.Optional[typing.Callable[[int, int, UserInterface.KeyboardModifiers], None]] = None
        self.__overlay_canvas_item = BitmapOverlayCanvasItem()
        self.add_canvas_item(self.__overlay_canvas_item)

    def close(self) -> None:
        self.on_drop_mime_data = None
        self.on_delete = None
        self.on_drag_pressed = None
        super().close()

    @property
    def active(self) -> bool:
        return self.__overlay_canvas_item.is_active

    @active.setter
    def active(self, value: bool) -> None:
        self.__overlay_canvas_item.is_active = value

    @property
    def focused(self) -> bool:
        return self.__overlay_canvas_item.focused

    def _set_focused(self, focused: bool) -> None:
        self.__overlay_canvas_item.is_focused = focused

    def update_sizing(self, new_sizing: CanvasItem.Sizing) -> None:
        super().update_sizing(new_sizing)
        self.__overlay_canvas_item.update_sizing(new_sizing)

    def drag_enter(self, mime_data: UserInterface.MimeData) -> str:
        self.__overlay_canvas_item.is_dropping = True
        return "ignore"

    def drag_leave(self) -> str:
        self.__overlay_canvas_item.is_dropping = False
        return "ignore"

    def drop(self, mime_data: UserInterface.MimeData, x: int, y: int) -> str:
        if callable(self.on_drop_mime_data):
            result = self.on_drop_mime_data(mime_data, x, y)
            if result:
                return result
        return super().drop(mime_data, x, y)

    def key_pressed(self, key: UserInterface.Key) -> bool:
        if key.is_delete:
            on_delete = self.on_delete
            if callable(on_delete):
                on_delete()
                return True
        return super().key_pressed(key)

    def mouse_pressed(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        self.__drag_start = Geometry.IntPoint(x=x, y=y)
        return True

    def mouse_released(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        self.__drag_start = None
        return True

    def mouse_position_changed(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        if self.__drag_start is not None and Geometry.distance(Geometry.FloatPoint(y, x), self.__drag_start.to_float_point()) > 2:
            self.__drag_start = None
            on_drag_pressed = self.on_drag_pressed
            if on_drag_pressed:
                on_drag_pressed(x, y, modifiers)
                return True
        return False


class ThumbnailCanvasItem(CanvasItem.CanvasItemComposition):
    """A canvas item that displays a thumbnail and allows dragging the thumbnail.

    To facilitate reusing existing canvas items, set the thumbnail source using set_thumbnail_source.

    This class keeps a bitmap and an overlay canvas item. The overlay canvas item is used for drag and drop events.

    Callers can set the on_drag, on_drop_mime_data, and on_delete callbacks to handle these events.
    """

    def __init__(self, ui: UserInterface.UserInterface, thumbnail_source: AbstractThumbnailSource, size: typing.Optional[Geometry.IntSize] = None) -> None:
        super().__init__()
        self.__ui = ui
        self.__thumbnail_source = thumbnail_source
        self.__thumbnail_size = size

        # define the callbacks
        self.on_drag: typing.Optional[typing.Callable[[UserInterface.MimeData, typing.Optional[_ImageDataType], int, int], None]] = None
        self.on_drop_mime_data: typing.Optional[typing.Callable[[UserInterface.MimeData, int, int], str]] = None
        self.on_delete: typing.Optional[typing.Callable[[], None]] = None

        # set up the initial bitmap and overlay canvas items.
        bitmap_overlay_canvas_item = BitmapOverlayCanvasItemComposition()
        bitmap_canvas_item = CanvasItem.BitmapCanvasItem(background_color="#CCC", border_color="#444")
        bitmap_overlay_canvas_item.insert_canvas_item(bitmap_overlay_canvas_item.canvas_items_count - 1, bitmap_canvas_item)
        if size is not None:
            bitmap_canvas_item.update_sizing(bitmap_canvas_item.sizing.with_fixed_size(size))
            bitmap_overlay_canvas_item.update_sizing(bitmap_overlay_canvas_item.sizing.with_fixed_size(size))
            for overlay_canvas_item in thumbnail_source.overlay_canvas_items:
                overlay_canvas_item.update_sizing(overlay_canvas_item.sizing.with_fixed_size(size))
        for overlay_canvas_item in thumbnail_source.overlay_canvas_items:
            bitmap_overlay_canvas_item.insert_canvas_item(bitmap_overlay_canvas_item.canvas_items_count - 1, overlay_canvas_item)

        # handle overlay drop callback by forwarding to the callback set by the caller.
        def drop_mime_data(mime_data: UserInterface.MimeData, x: int, y: int) -> str:
            if callable(self.on_drop_mime_data):
                return self.on_drop_mime_data(mime_data, x, y)
            return "ignore"

        # handle overlay drag callback by forwarding to the callback set by the caller.
        def delete() -> None:
            on_delete = self.on_delete
            if callable(on_delete):
                on_delete()

        # connect the handlers to the overlay canvas item.
        bitmap_overlay_canvas_item.on_drag_pressed = ReferenceCounting.weak_partial(ThumbnailCanvasItem.__drag_pressed, self)
        bitmap_overlay_canvas_item.on_drop_mime_data = drop_mime_data
        bitmap_overlay_canvas_item.on_delete = delete

        # store these for later use.
        self.__bitmap_canvas_item = bitmap_canvas_item
        self.__bitmap_overlay_canvas_item = bitmap_overlay_canvas_item

        # connect the thumbnail source data changed event to a handler and call the handler to initialize.
        self.__thumbnail_source.on_thumbnail_data_changed = ReferenceCounting.weak_partial(ThumbnailCanvasItem.__thumbnail_data_changed, self)
        self.__thumbnail_data_changed(self.__thumbnail_source.thumbnail_data)

        # add the overlay canvas item (with the bitmap canvas item) to this canvas item.
        self.add_canvas_item(bitmap_overlay_canvas_item)

    def __drag_pressed(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> None:
        # handle drag by forwarding to the callback set by the caller.
        on_drag = self.on_drag
        if callable(on_drag):
            mime_data = self.__ui.create_mime_data()
            valid, thumbnail = self.__thumbnail_source.populate_mime_data_for_drag(mime_data, Geometry.IntSize(width=80, height=80))
            if valid:
                on_drag(mime_data, thumbnail, x, y)

    def __thumbnail_data_changed(self, thumbnail_data: typing.Optional[_NDArray]) -> None:
        # update the bitmap canvas item with the new thumbnail data.
        self.__bitmap_canvas_item.rgba_bitmap_data = thumbnail_data

    @property
    def thumbnail_source(self) -> AbstractThumbnailSource:
        return self.__thumbnail_source

    def set_thumbnail_source(self, thumbnail_source: AbstractThumbnailSource) -> None:
        with self.batch_update():
            # reconfigure with the new thumbnail source.
            self.__thumbnail_source.close()
            self.__thumbnail_source = thumbnail_source
            if self.__thumbnail_size is not None:
                for overlay_canvas_item in thumbnail_source.overlay_canvas_items:
                    overlay_canvas_item.update_sizing(overlay_canvas_item.sizing.with_fixed_size(self.__thumbnail_size))
            while self.__bitmap_overlay_canvas_item.canvas_items_count > 1:
                self.__bitmap_overlay_canvas_item.remove_canvas_item(self.__bitmap_overlay_canvas_item.canvas_items[0])
            for overlay_canvas_item in thumbnail_source.overlay_canvas_items:
                self.__bitmap_overlay_canvas_item.insert_canvas_item(self.__bitmap_overlay_canvas_item.canvas_items_count - 1, overlay_canvas_item)
            self.__thumbnail_source.on_thumbnail_data_changed = ReferenceCounting.weak_partial(ThumbnailCanvasItem.__thumbnail_data_changed, self)
            self.__thumbnail_data_changed(self.__thumbnail_source.thumbnail_data)

    def close(self) -> None:
        self.__thumbnail_source.close()
        self.__thumbnail_source = typing.cast(typing.Any, None)
        self.on_drag = None
        self.on_drop_mime_data = None
        self.on_delete = None
        super().close()


class ThumbnailWidget(Widgets.CompositeWidgetBase):

    # when this widget is placed within a container, it will have no intrinsic size unless size is passed as a
    # parameter. for the case where the size parameter is unspecified, setting the size policy to expanding (in both
    # directions) tells the container that this widget would like to use all available space. however, in order for this
    # to take effect, the container hierarchy cannot utilize unbound stretches or else this widget will not expand. the
    # minimum size is present so that it always uses at least 32x32 pixels. the square canvas layout ensures that the
    # thumbnail area is always square and aligned to the top-left of the container.

    def __init__(self, ui: UserInterface.UserInterface, thumbnail_source: AbstractThumbnailSource,
                 size: typing.Optional[Geometry.IntSize] = None,
                 properties: typing.Optional[Persistence.PersistentDictType] = None) -> None:
        content_widget = ui.create_column_widget()
        super().__init__(content_widget)
        size = size or Geometry.IntSize(width=80, height=80)
        thumbnail_canvas_item = ThumbnailCanvasItem(ui, thumbnail_source, size)
        thumbnail_canvas_item.update_sizing(thumbnail_canvas_item.sizing.with_preferred_aspect_ratio(1.0))
        properties = properties or ({"height": size.height, "width": size.width} if size else dict())
        bitmap_canvas_widget = ui.create_canvas_widget(properties=properties)
        thumbnail_square_row = CanvasItem.CanvasItemComposition()
        thumbnail_square_row.layout = CanvasItem.CanvasItemRowLayout()
        thumbnail_square_row.add_canvas_item(thumbnail_canvas_item)
        thumbnail_square_row.add_stretch()
        thumbnail_square_column = CanvasItem.CanvasItemComposition()
        thumbnail_square_column.layout = CanvasItem.CanvasItemColumnLayout()
        thumbnail_square_column.add_canvas_item(thumbnail_square_row)
        thumbnail_square_column.add_stretch()
        bitmap_canvas_widget.canvas_item.add_canvas_item(thumbnail_square_column)
        content_widget.add(bitmap_canvas_widget)
        self.on_drop_mime_data: typing.Optional[typing.Callable[[UserInterface.MimeData, int, int], str]] = None
        self.on_drag: typing.Optional[typing.Callable[[UserInterface.MimeData, typing.Optional[_ImageDataType], int, int], None]] = None
        self.on_delete: typing.Optional[typing.Callable[[], None]] = None

        def drop_mime_data(mime_data: UserInterface.MimeData, x: int, y: int) -> str:
            if callable(self.on_drop_mime_data):
                return self.on_drop_mime_data(mime_data, x, y)
            return "ignore"

        def drag(mime_data: UserInterface.MimeData, thumbnail: typing.Optional[_NDArray], x: int, y: int) -> None:
            on_drag = self.on_drag
            if callable(on_drag):
                on_drag(mime_data, thumbnail, x, y)

        def delete() -> None:
            on_delete = self.on_delete
            if callable(on_delete):
                on_delete()

        thumbnail_canvas_item.on_drop_mime_data = drop_mime_data
        thumbnail_canvas_item.on_drag = drag
        thumbnail_canvas_item.on_delete = delete

    def close(self) -> None:
        self.on_drop_mime_data = None
        self.on_drag = None
        self.on_delete = None
        super().close()


class IsLiveOverlayCanvasItemComposer(CanvasItem.BaseComposer):
    def __init__(self, canvas_item: CanvasItem.AbstractCanvasItem, layout_sizing: CanvasItem.Sizing, cache: CanvasItem.ComposerCache, is_active: bool) -> None:
        super().__init__(canvas_item, layout_sizing, cache)
        self.__is_active = is_active

    def _repaint(self, drawing_context: DrawingContext.DrawingContext, canvas_bounds: Geometry.IntRect, composer_cache: CanvasItem.ComposerCache) -> None:
        is_active = self.__is_active
        with drawing_context.saver():
            drawing_context.translate(canvas_bounds.left, canvas_bounds.top)
            if is_active:
                with drawing_context.saver():
                    drawing_context.begin_path()
                    drawing_context.round_rect(2, 2, 6, 6, 3)
                    drawing_context.fill_style = "rgba(0, 255, 0, 0.80)"
                    drawing_context.fill()


class IsLiveOverlayCanvasItem(CanvasItem.AbstractCanvasItem):
    def __init__(self) -> None:
        super().__init__()
        self.__active = False

    @property
    def active(self) -> bool:
        return self.__active

    @active.setter
    def active(self, value: bool) -> None:
        if value != self.__active:
            self.__active = value
            self.update()

    def _get_composer(self, composer_cache: CanvasItem.ComposerCache) -> typing.Optional[CanvasItem.BaseComposer]:
        return IsLiveOverlayCanvasItemComposer(self, self.layout_sizing, composer_cache, self.__active)


class DataItemThumbnailSource(AbstractThumbnailSource):

    def __init__(self, ui: UserInterface.UserInterface, *,
                 display_item: typing.Optional[DisplayItem.DisplayItem] = None,
                 window: typing.Optional[Window.Window] = None,
                 overlay_canvas_items: typing.Optional[typing.Sequence[CanvasItem.AbstractCanvasItem]] = None) -> None:
        super().__init__()
        self.ui = ui
        self.__display_item: typing.Optional[DisplayItem.DisplayItem] = None
        self.__window = window
        self.__display_item_binding: typing.Optional[Binding.Binding] = None
        self.__thumbnail_source: typing.Optional[Thumbnails.ThumbnailSource] = None
        self.__thumbnail_updated_event_listener: typing.Optional[Event.EventListener] = None
        self.__display_item_changed_event: Event.EventListener | None = None

        self.is_live_overlay_canvas_item = IsLiveOverlayCanvasItem()

        self.overlay_canvas_items = [self.is_live_overlay_canvas_item] + (list(overlay_canvas_items) if overlay_canvas_items else [])

        if display_item:
            self.set_display_item(display_item)
        self.__update_display_item_task: typing.Optional[asyncio.Task[None]] = None

    def close(self) -> None:
        self.__detach_listeners()
        if self.__display_item_binding:
            self.__display_item_binding.close()
            self.__display_item_binding = None
        if self.__update_display_item_task:
            self.__update_display_item_task.cancel()
            self.__update_display_item_task = None
        super().close()

    def __detach_listeners(self) -> None:
        self.__thumbnail_updated_event_listener = None
        self.__thumbnail_source = None

    def __update_thumbnail(self) -> None:
        if self.__display_item:
            self._set_thumbnail_data(Thumbnails.ThumbnailManager().thumbnail_data_for_display_item(self.__display_item))
        else:
            self._set_thumbnail_data(None)
        if callable(self.on_thumbnail_data_changed):
            self.on_thumbnail_data_changed(self.thumbnail_data)

    def set_display_item(self, display_item: typing.Optional[DisplayItem.DisplayItem]) -> None:
        if self.__display_item != display_item:
            self.__detach_listeners()
            self.__display_item = display_item
            if display_item:
                self.__thumbnail_source = Thumbnails.ThumbnailManager().thumbnail_source_for_display_item(self.ui, display_item)
                self.__thumbnail_updated_event_listener = self.__thumbnail_source.thumbnail_updated_event.listen(self.__update_thumbnail)
                self.__display_item_changed_event = display_item.item_changed_event.listen(self.__handle_display_item_changed_event)
            else:
                self.__display_item_changed_event = None
            self.__update_thumbnail()
            self.__handle_display_item_changed_event()
            if self.__display_item_binding:
                self.__display_item_binding.update_source(display_item)

    def __handle_display_item_changed_event(self) -> None:
        if self.__display_item:
            self.is_live_overlay_canvas_item.active = self.__display_item.is_live
        else:
            self.is_live_overlay_canvas_item.active = False

    def populate_mime_data_for_drag(self, mime_data: UserInterface.MimeData, size: Geometry.IntSize) -> typing.Tuple[bool, typing.Optional[_NDArray]]:
        if self.__display_item:
            MimeTypes.mime_data_put_display_item(mime_data, self.__display_item)
            rgba_image_data = self.__thumbnail_source.thumbnail_data if self.__thumbnail_source else None
            thumbnail = Image.get_rgba_data_from_rgba(Image.scaled(Image.get_rgba_view_from_rgba_data(rgba_image_data), (80, 80))) if rgba_image_data is not None else None
            return True, thumbnail
        return False, None

    @property
    def display_item(self) -> typing.Optional[DisplayItem.DisplayItem]:
        return self.__display_item

    @display_item.setter
    def display_item(self, value: typing.Optional[DisplayItem.DisplayItem]) -> None:
        self.set_display_item(value)

    def bind_display_item(self, binding: Binding.Binding) -> None:
        if self.__display_item_binding:
            self.__display_item_binding.close()
            self.__display_item_binding = None
        self.display_item = binding.get_target_value()
        self.__display_item_binding = binding

        def update_display_item(display_item: DisplayItem.DisplayItem) -> None:
            if self.__window:
                async def update_display_item_() -> None:
                    with Process.audit("thumbnail_source.update_display_item"):
                        self.display_item = display_item
                        self.__update_display_item_task = None

                self.__update_display_item_task = self.__window.event_loop.create_task(update_display_item_())

        self.__display_item_binding.target_setter = update_display_item

    def unbind_display_item(self) -> None:
        if self.__display_item_binding:
            self.__display_item_binding.close()
            self.__display_item_binding = None


class DataItemReferenceThumbnailSource(DataItemThumbnailSource):
    """Used to track a data item referenced by a data item reference.

    Useful, for instance, for displaying a live update thumbnail that can be dragged to other locations."""

    def __init__(self, document_controller: DocumentController.DocumentController, data_item_reference: DocumentModel.DocumentModel.DataItemReference) -> None:
        data_item = data_item_reference.data_item
        document_model = document_controller.document_model
        display_item = document_model.get_display_item_for_data_item(data_item) if data_item else None
        super().__init__(document_controller.ui, display_item=display_item)

        async def async_data_item_changed() -> None:
            data_item = data_item_reference.data_item
            display_item = document_model.get_display_item_for_data_item(data_item) if data_item else None
            self.set_display_item(display_item)

        def data_item_changed() -> None:
            # this is a messy hack. the display item may not be available at the time this method is called; so delay
            # the call to set_display_item until the next event loop iteration.
            document_controller.event_loop.create_task(async_data_item_changed())

        self.__data_item_reference_changed_event_listener = data_item_reference.data_item_reference_changed_event.listen(data_item_changed)

    def close(self) -> None:
        self.__data_item_reference_changed_event_listener.close()
        self.__data_item_reference_changed_event_listener = typing.cast(typing.Any, None)
        super().close()


DataItemThumbnailWidget = ThumbnailWidget
